using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.OpenAccess;
using Telerik.OpenAccess.Metadata;

namespace AuditTrailExample
{
    public class AuditTrailExampleContext : OpenAccessContext
    {
        static MetadataContainer metadataContainer = new AuditTrailExampleMetadataSource().GetModel();
        static BackendConfiguration backendConfiguration = new BackendConfiguration()
        {
            Backend = "mssql"
        };

        private const string DbConnection = @"data source=.\sqlexpress;initial catalog=AuditTrailExampleDatabase;integrated security=True";

        public AuditTrailExampleContext()
            : base(DbConnection, backendConfiguration, metadataContainer)
        {
            //this.Events.Changed += ChangedEvent;
            this.Events.Added += AddedEvent;
            this.Events.Removed += RemovedEvent;
        }

        public AuditTrailExampleContext(bool forSchemaUpdate)
            : base(DbConnection, backendConfiguration, metadataContainer)
        {
            if (forSchemaUpdate == false)
            {
                this.Events.Changed += ChangedEvent;
                this.Events.Added += AddedEvent;
                this.Events.Removed += RemovedEvent;
            }
        }

        public IQueryable<Product> Products
        {
            get
            {
                return this.GetAll<Product>();
            }
        }

        public void UpdateSchema()
        {
            var handler = this.GetSchemaHandler();
            string script = null;
            try
            {
                script = handler.CreateUpdateDDLScript(null);
            }
            catch
            {
                bool throwException = false;
                try
                {
                    handler.CreateDatabase();
                    script = handler.CreateDDLScript();
                }
                catch
                {
                    throwException = true;
                }
                if (throwException)
                    throw;
            }

            if (string.IsNullOrEmpty(script) == false)
            {
                handler.ExecuteDDLScript(script);
            }
        }

        #region audit trail

        Dictionary<ObjectKey, Audit> audits = new Dictionary<ObjectKey, Audit>();

        private void CollectChanges(ChangeEventArgs e)
        {
            var key = CreateObjectKey(e.PersistentObject);
            Audit auditObject = null;

            if (this.audits.TryGetValue(key, out auditObject) == false)
            {
                auditObject = CreateAudit(key, Audit.AuditAction.Update);
                this.audits.Add(key, auditObject);
            }

            auditObject.OldData += string.Format("'{0}'- Old:'{1}'", e.PropertyName, e.OldValue != null ? e.OldValue.ToString() : string.Empty);
            auditObject.NewData += string.Format("'{0}'- New:'{1}'", e.PropertyName, e.NewValue.ToString());
        }

        internal void AddChangedAudit(ObjectKey key)
        {
            Audit audit = null;
            if (this.audits.TryGetValue(key, out audit))
            {
                audit.AuditDate = DateTime.Now;
                this.audits.Remove(key);
                this.Add(audit);
            }
        }

        void ChangedEvent(object sender, ChangeEventArgs e)
        {
            if (e.PersistentObject.GetType() == typeof(Audit))
                return;

            CollectChanges(e);
        }

        void AddedEvent(object sender, AddEventArgs e)
        {
            if (e.PersistentObject.GetType() == typeof(Audit))
                return;
            var audit = CreateAudit(CreateObjectKey(e.PersistentObject),Audit.AuditAction.Insert);
            this.Add(audit);
        }

        void RemovedEvent(object sender, RemoveEventArgs e)
        {
            if (e.PersistentObject.GetType() == typeof(Audit))
                return;
            var audit = CreateAudit(CreateObjectKey(e.PersistentObject),Audit.AuditAction.Delete);
            this.Add(audit);
        }

        private Audit CreateAudit(ObjectKey key,Audit.AuditAction action)
        {
            var audit = new Audit() { ObjectId = key.ToString(), Action = action,ClassName = key.TypeName,AuditDate = DateTime.Now};

            var pt = this.Metadata.PersistentTypes.Where(pts => pts.FullName.Equals(key.TypeName)).Single();
            if (pt != null && pt.Table != null)
                audit.TableName = pt.Table.Name;

            return audit;
        }

        
        #endregion
    }
}
